function [mute_result, model, ignoredReactions] = mute(model,translationEfficiency,geneExpression,proteinWeight,enzymeKinetics,geneNames,proteinWeightFull,geneNamesWeight,varargin)

%% MUTE Metabolic Flux Distribution by Translational Efficiency and Enzyme Kinetics
%   
%   [mute_result] = mute(model,translationEfficiency,geneExpression,enzymeKinetics,...parameters....)
%
%   Inputs
%   model                      COBRA model structure
%   translationalEfficiency    Translational efficiency for mRNAs. Each
%                              element i corresponds to the translational
%                              efficiency of mRNA i. Indexed in same order
%                              as rxns field of 'model'.
%   geneExpression             Gene expression values for mRNAs. Each
%                              element i corresponds to the mean measured 
%                              expression value of mRNA i. Indexed in same 
%                              order as rxns field of 'model'.
%   enzymeKinetics             Kinetic parameters for all enzymes. Each
%                              element i corresponds to the kinetic 
%                              parameter of enzyme i. Indexed in same order
%                              as rxns field of 'model'.
%   proteinWeight              Molecular weight of all proteins. Each
%                              element i corresponds to the molecular
%                              weight of protein i. Indexed in same order
%                              as rxns field of 'model'.
%
%   Parameters
%   'fraction_enzymes'         The fraction of cell dry weight occupied by 
%                              metab    olic enzymes. Default value = 0.27.
%   'cell_dry_weight'          The dry weight of the cell, measured in
%                              grams. Default value = 280*10e-15.
%   'cell_volume'              Volume of the modelled cell, liters.
%   Outputs
%   mute_result                Solution structure.
%   mute_model                 Modified cobra model structure.
%
%
%   Eivind B. Drejer            13.03.2015

%% Parse optional parameters
p = inputParser;

p.addParamValue('fraction_enzymes',0.27);
p.addParamValue('cell_dry_weight',280e-15);
p.addParamValue('cell_volume',0.67e-15);
p.addParamValue('median_prot_copy_nmb',526);
p.addParamValue('prot_pr_mrna',540);
p.parse(varargin{:});

%%  Optimize protein concentrations to match corresponding mRNA level, scaled by translation efficiency parameter
%   Total enzyme mass must be less than or equal to the fraction of dry
%   cell mass dedicated to enzymes.

unaccountedMass = 0;
for i = 1:length(model.genes)
    if ~ismember(model.genes(i),geneNames)
        unaccountedMass = sum([unaccountedMass proteinWeightFull{find(strcmp(model.genes(i),geneNamesWeight)==1)}]);
    end
end
avogadro_number = 6.02214078e23;
unaccountedMass = p.Results.median_prot_copy_nmb*unaccountedMass/avogadro_number;
enzyme_upper_limit = ((p.Results.fraction_enzymes * p.Results.cell_dry_weight) - unaccountedMass);
miliScaling = 1000;

Gmodel.A = proteinWeight'/miliScaling;
Gmodel.A = [Gmodel.A; 1000*eye(length(proteinWeight))];
Gmodel.A = sparse(Gmodel.A);
Gmodel.obj = -2*translationEfficiency.*geneExpression*p.Results.prot_pr_mrna/avogadro_number*miliScaling/p.Results.cell_volume;
Gmodel.Q = sparse(eye(length(geneExpression)));
Gmodel.modelsense = 'min';
% Convert total enzyme mass restriction from g to g/L
Gmodel.rhs = enzyme_upper_limit/p.Results.cell_volume;
Gmodel.rhs = [Gmodel.rhs; 20*ones(length(proteinWeight),1)];
Gmodel.sense  = '<';
Gmodel.objcon = sum((p.Results.prot_pr_mrna*geneExpression.*translationEfficiency/avogadro_number*miliScaling/p.Results.cell_volume).^2);
Gmodel.varnames = geneNames;
params.OutputFlag = 1;
params.TuneOutput = 1;
%params.ScaleFlag = 1;
%params.FeasibilityTol = 1e-9;
%params.BarConvTol = 0.0;
params.BarHomogeneous = 1;
protein_conc = gurobi(Gmodel, params); 
%protein_conc.x = protein_conc.x*1000;
% Temporary fix to scaling problems

%% Assign effective enzyme concentrations to all reactions in COBRA model
% mean_proteinConc = mean(protein_conc.x);
model.enzConc = ones(length(model.rxns),1);
for i = 1:length(model.rxns)
    node = parse_string(model.grRules{i,1});
    if ~node.NULL
        test = recurseReaction(node,protein_conc.x,geneNames);
        if ~isnan(test)
            model.enzConc(i,1) = test;
        else
            model.enzConc(i,1) = NaN;
        end
    else
        model.enzConc(i,1) = NaN;
    end
end

%% Assign enzyme turnover numbers to all reactions in COBRA model
% mean_turnover = 0;
% for i = 1:length(enzymeKinetics)
%    mean_turnover = mean_turnover + enzymeKinetics{i,2};
% end
% mean_turnover = mean_turnover / length(enzymeKinetics);

model.enzKinetics = ones(length(model.rxns),1);
for i = 1:length(model.enzKinetics)
    model.enzKinetics(i) = NaN;
end
for i = 1:length(enzymeKinetics)
    index = find(model.rxnGeneMat(:,find(strcmp(model.genes,geneNames{i,1}))) ==1);
    for j = 1:length(index)
        if ~isempty(index) && ~isnan(model.enzConc(index(j),1)) 
            model.enzKinetics(index) = enzymeKinetics(i,1);
        end
    end
end

for i = 1:length(model.enzKinetics)
   if isnan(model.enzKinetics(i,1))
      model.enzConc(i,1) = NaN; 
   end
end

clearvars Gmodel index

%% Create Gurobi model structure for flux optimization

Gmodel.A = model.S;
Gmodel.Q = sparse(eye(length(model.rxns)));
for i = 1:length(model.rxns)
   if isnan(model.enzConc(i,1))
       Gmodel.Q(i,i) = 0;
   end
end
%Gmodel.Q = sparse(eye(length(model.rxns(find(~isnan(model.enzConc))))));
% Convert flux unit from mmol/(L*s) to mmol/(gDW*h)
Gmodel.obj = zeros(length(model.rxns),1);
for i = 1:length(model.rxns)
   if ~isnan(model.enzConc(i,1))
       Gmodel.obj(i) = -2*(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1))*(3600*model.enzKinetics(i));
   end
end
%Gmodel.obj = -2*(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(find(~isnan(model.enzConc)))).*(3600*model.enzKinetics(find(~isnan(model.enzKinetics))));
Gmodel.modelsense = 'min';
Gmodel.rhs = zeros(length(model.mets),1);
Gmodel.sense = '=';

% Find growth rate predicted by FBA
fba_result = optimizeCbModel(model);

% Set minimum biomass reaction flux to some fraction of "optimal" flux
Gmodel.lb = model.lb;
Gmodel.ub = model.ub;
exc_rxns = findExcRxns(model);
testmodel = model;
ignoredReactions = '';
for i = 1:length(model.rxns)
   isSink = strfind(model.rxnNames{i},'Sink');
   if ((model.rev(i)) && (i ~= find(model.c~=0)) && ~exc_rxns(i) && (~isnan(model.enzKinetics(i))))
       testmodel.lb(i) = max(-50,-(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
       testmodel.ub(i) = min(50,(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
       testresult = optimizeCbModel(testmodel);
       if testresult.f >= fba_result.f*0.1
           Gmodel.lb(i) = max(-50,-(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
           Gmodel.ub(i) = min(50,(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
           model.ub(i) = min(50,(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
           model.lb(i) = max(-50,-(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
       else
           testmodel.lb(i) = max(model.lb(i),-50);
           testmodel.ub(i) = min(model.ub(i),50);
           model.lb(i) = max(model.lb(i),-50);
           model.ub(i) = min(model.ub(i),50);
           ignoredReactions = [ignoredReactions model.rxns(i)];
       end
   elseif ((~model.rev(i)) && (i ~= find(model.c~=0)) && ~exc_rxns(i) && (~isnan(model.enzKinetics(i))))
       testmodel.ub(i) = min(50,(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
       testmodel.lb(i) = 0;
       testresult = optimizeCbModel(testmodel);
       if testresult.f >= fba_result.f*0.1
           Gmodel.lb(i) = 0;
           Gmodel.ub(i) = min(50,(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
           model.ub(i) = min(50,(p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(i,1)).*(3600*model.enzKinetics(i,1)));
           model.lb(i) = 0;
       else
           testmodel.lb(i) = max(model.lb(i),-50);
           testmodel.ub(i) = min(model.ub(i),50);
           model.ub(i) = min(model.ub(i),50);
           model.lb(i) = max(model.lb(i),-50);
           ignoredReactions = [ignoredReactions model.rxns(i)];
       end
   elseif ((~model.rev(i)) && (i ~= find(model.c~=0)) && exc_rxns(i) && ~isempty(isSink))
       testmodel.ub(i) = min(20,model.ub(i));
       testmodel.lb(i) = 0;
       testresult = optimizeCbModel(testmodel);
       if testresult.f >= fba_result.f*0.1
           Gmodel.lb(i) = 0;
           Gmodel.ub(i) = min(20,model.ub(i));
           model.lb(i) = 0;
           model.ub(i) = min(20,model.ub(i));
       else
           testmodel.ub(i) = model.ub(i);
           testmodel.lb(i) = model.lb(i);
           model.ub(i) = min(model.ub(i),20);
           model.lb(i) = 0;
       end
   elseif ((model.rev(i)) && (i ~= find(model.c~=0)) && exc_rxns(i) && ~isempty(isSink))
       testmodel.ub(i) = min(20,model.ub(i));
       testmodel.lb(i) = max(-20,model.lb(i));
       testresult = optimizeCbModel(testmodel);
       if testresult.f >= fba_result.f*0.1
           Gmodel.lb(i) = max(-20,model.lb(i));
           Gmodel.ub(i) = min(20,model.ub(i));
           model.lb(i) = max(-20,model.lb(i));
           model.ub(i) = min(20,model.ub(i));
       else
           testmodel.ub(i) = model.ub(i);
           testmodel.lb(i) = model.lb(i);
           model.ub(i) = min(model.ub(i),20);
       end
   elseif model.rev(i)
      testmodel.ub(i) = min(model.ub(i),20);
      testmodel.lb(i) = max(model.lb(i),-20);
      model.ub(i) = min(model.ub(i),20);
      model.lb(i) = max(model.lb(i),-20);
   else
      testmodel.ub(i) = min(model.ub(i),20);
      model.ub(i) = min(model.ub(i),20); 
   end
end


%Gmodel.lb(find(model.c~=0)) = 0.1*fba_result.f;
%model.lb(find(model.c~=0)) = 0.1*fba_result.f;

% Objective offset - sum of enzyme concentrations multiplied by kinetic
% parameters
% Convert flux unit from mmol/(L*s) to mmol/(gDW*h)
Gmodel.objcon = sum(((p.Results.cell_volume/p.Results.cell_dry_weight*model.enzConc(find(~isnan(model.enzConc)))).*(3600*model.enzKinetics(find(~isnan(model.enzKinetics))))).^2);
Gmodel.varnames = model.rxns;
params.OutputFlag =0;
params.TuneOutput = 1;
params.BarIterLimit = 10000;
mute_result = gurobi(Gmodel,params);


